/*  Copyright (c) MIPoPS. All Rights Reserved.
 *
 *  Use of this source code is governed by a Unlicense license that can
 *  be found in the LICENSE file in the root of the source tree.
 */

//---------------------------------------------------------------------------
#include "Common/Core.h"
#include "ZenLib/Ztring.h" // TODO: remove that
using namespace ZenLib;
//---------------------------------------------------------------------------


//***************************************************************************
// Internal
//***************************************************************************

class file
{
public:
    MediaInfo MI;
    std::vector<MediaInfo_Event_DvDif_Analysis_Frame_0*> PerFrame;
    float FrameRate;

    file(const String& FileName);
    ~file();

    void AddFrame(const MediaInfo_Event_DvDif_Analysis_Frame_0* FrameData);
};

//***************************************************************************
// Callback
//***************************************************************************

void __stdcall Event_CallBackFunction(unsigned char* Data_Content, size_t Data_Size, void* UserHandler_Void)
{
    //Retrieving UserHandler
    file*                               UserHandler = (file*)UserHandler_Void;
    struct MediaInfo_Event_Generic*     Event_Generic = (struct MediaInfo_Event_Generic*) Data_Content;
    unsigned char                       ParserID;
    unsigned short                      EventID;
    unsigned char                       EventVersion;

    //Integrity test
    if (Data_Size < 4)
        return; //There is a problem

    //Retrieving EventID
    ParserID = (unsigned char)((Event_Generic->EventCode & 0xFF000000) >> 24);
    EventID = (unsigned short)((Event_Generic->EventCode & 0x00FFFF00) >> 8);
    EventVersion = (unsigned char)(Event_Generic->EventCode & 0x000000FF);
    switch (ParserID)
    {
    case MediaInfo_Parser_DvDif:
        switch (EventID)
        {
        case MediaInfo_Event_DvDif_Analysis_Frame: if (EventVersion == 0 && Data_Size == sizeof(struct MediaInfo_Event_DvDif_Analysis_Frame_0)) UserHandler->AddFrame((MediaInfo_Event_DvDif_Analysis_Frame_0*)Event_Generic); break;
        }
        break;
    }
}

//***************************************************************************
// Constructor/Destructor
//***************************************************************************

//---------------------------------------------------------------------------
file::file(const String& FileName)
{
    MI.Option(__T("File_Event_CallBackFunction"), __T("CallBack=memory://") + Ztring::ToZtring((size_t)&Event_CallBackFunction) + __T(";UserHandler=memory://") + Ztring::ToZtring((size_t)this));
    MI.Option(__T("File_DvDif_Analysis"), __T("1"));
    MI.Open(FileName);

    // Filing some info
    FrameRate = Ztring(MI.Get(Stream_Video, 0, __T("FrameRate_Original"))).To_float32();
    if (!FrameRate)
        FrameRate = Ztring(MI.Get(Stream_Video, 0, __T("FrameRate"))).To_float32();
    if (!FrameRate)
        FrameRate = float(30 / 1.001); // Default if no frame rate available
}

//---------------------------------------------------------------------------
file::~file()
{
    for (auto& Frame : PerFrame)
        if (Frame->Errors)
            delete[] Frame->Errors;
}

//***************************************************************************
// Helpers
//***************************************************************************

//---------------------------------------------------------------------------
void file::AddFrame(const MediaInfo_Event_DvDif_Analysis_Frame_0* FrameData)
{
    MediaInfo_Event_DvDif_Analysis_Frame_0* ToPush = new MediaInfo_Event_DvDif_Analysis_Frame_0();
    std::memcpy(ToPush, FrameData, sizeof(MediaInfo_Event_DvDif_Analysis_Frame_0));
    if (FrameData->Errors)
    {
        size_t SizeToCopy = std::strlen(FrameData->Errors) + 1;
        ToPush->Errors = new char[SizeToCopy];
        std::memcpy(ToPush->Errors, FrameData->Errors, SizeToCopy);
    }
    PerFrame.push_back(ToPush);
}

//***************************************************************************
// Constructor/Destructor
//***************************************************************************

//---------------------------------------------------------------------------
Core::Core()
{
    MediaInfo::Option_Static(__T("ParseSpeed"), __T("1.000"));
    Xml=false;
}

Core::~Core()
{
    PerFile_Clear();
}

//***************************************************************************
// Process
//***************************************************************************

//---------------------------------------------------------------------------
void Core::Process()
{
    PerFile_Clear();
    
    PerFile.reserve(Inputs.size());
    for (const auto& Input : Inputs)
    {
        PerFile.push_back(new file(Input));
    }
}

//---------------------------------------------------------------------------
float Core::State ()
{
    size_t Total = 0;
    for (const auto& File : PerFile)
    {
        Total += File->MI.State_Get();
    }
    return (((float)Total)/PerFile.size()/10000);
}

//***************************************************************************
// Output
//***************************************************************************

//---------------------------------------------------------------------------
string Core::OutputXml()
{
    string Text;

    // XML header
    Text += "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n";
    Text += "<dvrescue xmlns=\"https://mediaarea.net/dvrescue\" version=\"1.0\">\n";

    for (const auto& File : PerFile)
    {
        if (File->PerFrame.empty())
            continue; // Show the file only if there is some DV content

        // Media header
        Text += "\t<media ref=\"" + Ztring(File->MI.Get(Stream_General, 0, __T("CompleteName"))).To_UTF8() + "\">\n";
        Text += "\t\t<frames>\n";

        // By Frame - For each line
        auto FrameNumber_Max = File->PerFrame.size() - 1;
        for (const auto& Frame : File->PerFrame)
        {
            auto FrameNumber = &Frame - &*File->PerFrame.begin();

            if (Frame->Errors || FrameNumber==0 || FrameNumber + 1 == FrameNumber_Max)
            {
                auto TimeStamp = FrameNumber / File->FrameRate;

                Text += "\t\t\t<frame"
                    " n=\"" + to_string(FrameNumber) + "\""
                    " pts=\"" + to_string(TimeStamp) + "\"";


                auto TimeCode_Seconds = (Frame->TimeCode >> 8) & 0x7FFFF;
                if (TimeCode_Seconds != 0x7FFFF)
                {
                    string TimeCode_String = "00:00:00:00";
                    TimeCode_String[0] += TimeCode_Seconds / 36000; TimeCode_Seconds %= 36000;
                    TimeCode_String[1] += TimeCode_Seconds / 3600; TimeCode_Seconds %= 3600;
                    TimeCode_String[3] += TimeCode_Seconds / 600; TimeCode_Seconds %= 600;
                    TimeCode_String[4] += TimeCode_Seconds / 60; TimeCode_Seconds %= 60;
                    TimeCode_String[6] += TimeCode_Seconds / 10; TimeCode_Seconds %= 10;
                    TimeCode_String[7] += TimeCode_Seconds;
                    if (Frame->TimeCode & 0x00000080)
                        TimeCode_String[8] = ';';
                    auto TimeCode_Frames = Frame->TimeCode & 0x3F;
                    TimeCode_String[9] += TimeCode_Frames / 10;
                    TimeCode_String[10] += TimeCode_Frames % 10;
                    Text += " timecode=\"" + TimeCode_String + "\"";
                }

                //RecDate/RecTime
                string RecDateTime_String;
                auto RecDateTime_Years = (Frame->RecordedDateTime1 >> 17) & 0x7F;
                if (RecDateTime_Years != 0x7F)
                {
                    auto RecDateTime_Months = (Frame->RecordedDateTime2 >> 12) & 0x0F;
                    auto RecDateTime_Days = (Frame->RecordedDateTime2 >> 8) & 0x1F;
                    RecDateTime_String = "2000-00-00";
                    if (RecDateTime_Years >= 70) // Arbitrary decided
                    {
                        RecDateTime_String[0] = '1';
                        RecDateTime_String[1] = '9';
                    }
                    RecDateTime_String[2] += RecDateTime_Years / 10;
                    RecDateTime_String[3] += RecDateTime_Years % 10;
                    RecDateTime_String[5] += RecDateTime_Months / 10;
                    RecDateTime_String[6] += RecDateTime_Months % 10;
                    RecDateTime_String[8] += RecDateTime_Days / 10;
                    RecDateTime_String[9] += RecDateTime_Days % 10;
                }
                int32u RecDateTime_Seconds = Frame->RecordedDateTime1 & 0x1FFFF;
                if (RecDateTime_Seconds != 0x1FFFF)
                {
                    string RecTime_String = "00:00:00";
                    RecTime_String[0] += RecDateTime_Seconds / 36000; RecDateTime_Seconds %= 36000;
                    RecTime_String[1] += RecDateTime_Seconds / 3600; RecDateTime_Seconds %= 3600;
                    RecTime_String[3] += RecDateTime_Seconds / 600; RecDateTime_Seconds %= 600;
                    RecTime_String[4] += RecDateTime_Seconds / 60; RecDateTime_Seconds %= 60;
                    RecTime_String[6] += RecDateTime_Seconds / 10; RecDateTime_Seconds %= 10;
                    RecTime_String[7] += RecDateTime_Seconds;
                    int32u RecTime_Frames = Frame->RecordedDateTime2 & 0x7F;
                    if (RecTime_Frames != 0x7F)
                    {
                        /*
                        int32u RecDateTime_Milliseconds = float64_int32s(RecTime_Frames * 1000 / File->FrameRate);
                        RecTime_String += ".000";
                        RecTime_String[9] += RecDateTime_Milliseconds / 100; RecDateTime_Milliseconds %= 100;
                        RecTime_String[10] += RecDateTime_Milliseconds % 100; RecDateTime_Milliseconds %= 10;
                        RecTime_String[11] += RecDateTime_Milliseconds % 10;
                        */
                        RecTime_String += ":00";
                        RecTime_String[9] += RecTime_Frames / 10;
                        RecTime_String[10] += RecTime_Frames % 10;
                    }
                    if (!RecDateTime_String.empty())
                        RecDateTime_String += ' ';
                    RecDateTime_String += RecTime_String;
                }
                if (!RecDateTime_String.empty())
                    Text += " recdatetime=\"" + RecDateTime_String + "\"";

                //Start
                if (Frame->RecordedDateTime1&(1 << 29))
                    Text += " recstart=\"1\"";

                //End
                if (Frame->RecordedDateTime1&(1 << 28))
                    Text += " recend=\"1\"";

                Text += "/>\n";
            }
        }
            
        // Media footer
        Text += "\t\t</frames>\n";
        Text += "\t</media>\n";
    }

    // XML footer
    Text += "</dvrescue>\n";

    return Text;
}

//***************************************************************************
// PerFile
//***************************************************************************

//---------------------------------------------------------------------------
void Core::PerFile_Clear()
{
    for (const auto& File : PerFile)
    {
        delete File;
    }
    PerFile.clear();
}
